﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;

public enum GustState { Ready, AirbornLockout, DashWait, Wait };

public class PlayerMovement : MonoBehaviour {
	
	//don't forget to make variables static during your final code pass
	
	//----Movement Variables--------
	public Transform target;

	[Header("Movement Configuration")]
	// the 'normal' movement speed. because the movement speed is interpolated during GUST, we need to store the 
	//  normal speed it interpolates back to. that's this.
	[SerializeField]
	private float baseMoveSpeed = 9.5f;
	[SerializeField]
	private float rotSpeed = 5.0f;
	// used to give kicked boxes velocity.
	private Vector3 lastVelocity = new Vector3();

	// the real, frame to frame movement speed. this is volatile and is changed and rewritten by GUST
	//  interpolation. therefore, most of the movement code reads this value to decide how fast to go.
	//  GUST writes to it to create the boost and wind-down.
	private float appliedMoveSpeed;
	
	//----Jump Variables--------
	[Header("Jump Configuration")]
	[SerializeField]
	private float gravity = -11.0f;
	[SerializeField]
	private float gravMultiplier = 5.0f;
	[SerializeField]
	private float terminalVelocity = -20.0f;
	// how high the character can jump when on the ground.
	[SerializeField]
	private float groundJumpHeight = 0;

	// run time frame to frame tracker for the vertical speed of the player.
	//  subject to gravity and jumping.
	private float vertSpeed;
	// expected component on player object that facilitates jumping. provides
	//  interface for starting jumps and ending them early.
	private Jumper jumper;

	//----Gust Variables------------
	[Header("Gust Configuration")]
	// the amount of seconds spent waiting for gust to recharge, not including the gust wind-down.
	[SerializeField] 
	private float gustCooldown = 2f;
	// the amount of time the gust boost stays at full speed for.
	[SerializeField] 
	private float gustGroundBoostDuration = 1f;
	// the amount to time that the gust takes to interpolate from the boost speed back to the normal speed
	//  after the full speed duration has finished.
	[SerializeField] 
	private float gustGroundBoostWindDown = 2f;
	// how fast is the boost at full speed, in terms of the base speed?
	[SerializeField]
	private float gustGroundMovespeedMultiplier = 2f;
	// how high the gust-associated jump is.
	[SerializeField]
	private float gustJumpHeight = 0;

	// frame to frame state of the gust. used to track whether ready, in cooldown, airborn, ...
	private GustState gustState = GustState.Ready;
	// used to keep track of what part of the gust we're in. also used for gust wind-down interpolation.
	private Timer gustCooldownTimer;
	
	//----Particle Variables--------
	[Header("Attached Particle System")]
	public ParticleSystem kickFlash;
	private float dustSpeedCutoff = 6.0f; //minimum speed to kick up dust
	ParticleSystem dustSystem;
	
	//----Collision Variables--------
	private ControllerColliderHit contact;
	private CharacterController charController;
	//private Animator animator;
	
	void Start() {
		appliedMoveSpeed = baseMoveSpeed;

		//Application.targetFrameRate = 30;
		//QualitySettings.vSyncCount = 0;

		Utilities.RequireFindComponent(ref charController, gameObject);

		Utilities.RequireFindComponent(ref jumper, gameObject);
		jumper.SetGravity(gravity*gravMultiplier);

		Utilities.RequireFindComponentInChildren(ref dustSystem, gameObject);

		gustCooldownTimer = new Timer();

		//animator = GetComponent<Animator>();
	}

	void Update() {
		// placed here so that adjusting gravity while fiddling with editor will also update
		//  jumper gravity. can be removed in release IF gravity NEVER changes.
		jumper.SetGravity(gravity*gravMultiplier);

		// handles frame to frame implementation of gust ability.
		UpdateGust();

		// start with zero and add movement components progressively
		Vector3 movement = Vector3.zero;

		//get inputs
		float horInput = Input.GetAxis("xb_leftstick_x");
		float vertInput = -Input.GetAxis("xb_leftstick_y");
			
		if (horInput != 0 || vertInput != 0) {
			//x z movement transformed relative to target
			movement.x = horInput * appliedMoveSpeed;
			movement.z = vertInput * appliedMoveSpeed;
			movement = Vector3.ClampMagnitude(movement, appliedMoveSpeed);

			Quaternion tmp = target.rotation;
			target.eulerAngles = new Vector3(0, target.eulerAngles.y, 0);
			movement = target.TransformDirection(movement);
			target.rotation = tmp;

			// face movement direction
			Quaternion direction = Quaternion.LookRotation(movement, Vector3.up);
			transform.rotation = Quaternion.Lerp(transform.rotation, direction, rotSpeed * Time.deltaTime);
		}
		//animator.SetFloat("Speed", movement.sqrMagnitude);

		if (charController.isGrounded) {
			// In debug, jumper draws the arcs that are jumped.
			// All this function does is tell it the jump ended and that
			//  it can stop drawing the arc.
			jumper.HitGround();
			//only jump if grounded
			if(Input.GetButtonDown("xb_a")) { // JUMP
				// start a jump to hight: groundJumpHeight
				jumper.Jump(groundJumpHeight);
				// let the jumper tell us how the jump affects the vertical speed we already calculated.
				vertSpeed = jumper.ApplyToVerticalSpeed(vertSpeed);
			}
			else 
			{
				//vertSpeed was set to minVel here.
				//animator.SetBool("Jumping", false);
			}

			if (Input.GetButtonDown("xb_x") && gustState == GustState.Ready) { // GUST
				StartGust();
			}
		}
		else {
			if (Input.GetButtonUp("xb_a")) // End JUMP
			{ 
				// ends a jump early when button is released soon.
				jumper.ReleaseJump(); 
			}

			if (Input.GetButtonDown("xb_x") && gustState == GustState.Ready) { // GUST
				StartGust();
			}
			 
			// falling
			vertSpeed = Mathf.Max(vertSpeed + gravity * gravMultiplier * Time.deltaTime, terminalVelocity);
			// let the jumper tell us how the jump affects the vertical speed we already calculated.
			vertSpeed = jumper.ApplyToVerticalSpeed(vertSpeed);

			//take another look at this when adding animations
			if (contact == null ) {
				//do nothing
			}
			else {
				//not right at level start?
				//animator.SetBool("Jumping", true);
			}
			
			// The character can be in a situation where the center of the capsule collider is 
			// 	off of a platform, but a part of the lower sphere is. This codes causes the 
			//	character to SLIP off the platform in that situation, by apply more velocity
			//	downward-ish (from contact.normal), scaling it by appliedMoveSpeed, and adding it to movement.
			// NOTE: I tore out several lines related to this. Getting it to work again will
			//  require some work. Used removed variable minVel or minVert?
#if false
			if (slipOffEdges && charController.isGrounded) {
				if (Vector3.Dot(movement, contact.normal) < 0)
					{ movement = contact.normal * appliedMoveSpeed; }
				else
					{ movement += contact.normal * appliedMoveSpeed; }
			}
#endif
		}

		//turn on/off dust trails
		if (movement.magnitude < dustSpeedCutoff || !charController.isGrounded)
		{ StopEmit(); }
		else
		{ Emit(); }

		//add everything together, move player
		//movement.y = vertSpeed;
		movement *= Time.deltaTime;
		// in order to model the parabola correctly, the actual offset needs to 
		//  take gravity into account, let jumper do this for us.
		movement.y = jumper.OffsetFromVelocity(vertSpeed);
		lastVelocity = movement;
		charController.Move(movement);
	}

	// Collision code
	void OnControllerColliderHit(ControllerColliderHit hit) {
		contact = hit; //save for later
		if (hit.gameObject.tag == "KickObject") {
			//If player kicks crate, pass variables and create a kick particle effect.
			Vector3 hitHeading = (hit.transform.position - transform.position);
			Vector3 thisCollision = (hit.transform.position);
			BoxKick boxKick = hit.transform.gameObject.GetComponent<BoxKick>();
			boxKick.CalculateKick(hitHeading, lastVelocity);
			ParticleSystem thisKickFlash = Instantiate(kickFlash, thisCollision, transform.rotation);
		}
	}
	
	//these two functions are how the new particle system turns on/off
	void Emit()
	{ 
		ParticleSystem.EmissionModule em = dustSystem.emission;
		em.enabled = true;
	}
	
	void StopEmit()
	{ 
		ParticleSystem.EmissionModule em = dustSystem.emission;
		em.enabled = false;
	}

	// Call to start a gust, returns true if Gust was started, returns false if gust was not
	//  started. (ie. on cooldown)
	bool StartGust()
	{
		if (gustState != GustState.Ready)
		{ return false; }

		if (charController.isGrounded)
		{
			gustState = GustState.DashWait;
			
			gustCooldownTimer.Start(gustGroundBoostDuration+gustCooldown);
			// alternate logic:
			//gustCooldownTimer.Start(Mathf.Max(gustGroundBoostDuration,gustCooldown));
			appliedMoveSpeed = baseMoveSpeed*gustGroundMovespeedMultiplier;
		}
		else
		{
			gustState = GustState.AirbornLockout;
			jumper.Jump(gustJumpHeight);
		}
		return true;

	}

	void UpdateGust()
	{
		switch (gustState)
		{
			case GustState.Ready: {
				/* 
					We're waiting for the button press to start a Gust. 
					Do Nothing.
				*/
			 } break;
			case GustState.DashWait: { 
				float totalTime = 0f;
				if (gustCooldownTimer.UpdateAndCheck(ref totalTime))
				{ 
					// if the ground dash is completely finished, including cooldown,
					//  reset and unlock it for next use.
					gustState = GustState.Ready; 
					appliedMoveSpeed = baseMoveSpeed;
				}
				else
				{
					// if ground dash isn't done, manifest the ability.
					float time = gustCooldownTimer.Elapsed();
					if (time < gustGroundBoostDuration)
					{ 
						/* 
							Speed boost was applied when dash began, waiting for
							 full speed duration to end.
							Do Nothing.
						*/
					}
					else if (time < gustGroundBoostWindDown + gustGroundBoostDuration)
					{
						/* 
							If time is not < boostDuration, then this catches if it will still be in the windDownPeriod.
							 In that case, we're going to interpolate the movement speed down to normal.
						*/
						float timeRatio = (time-gustGroundBoostDuration)/gustGroundBoostWindDown;
						appliedMoveSpeed = Mathf.Lerp(
							baseMoveSpeed*gustGroundMovespeedMultiplier,
							baseMoveSpeed, timeRatio);
					}
					else
					{ 
						/* 
							If we're past the wind-down period, reset the movespeed to make sure the 
							 boost ends as it should.
							Resetting the GustState is done at 
							 if (gustCooldownTimer.UpdateAndCheck(ref totalTime))
						*/
						appliedMoveSpeed = baseMoveSpeed;
					}
				}
			} break;
			case GustState.AirbornLockout: 
			{ 
				/*
					An air gust set us to wait to hit the ground. 
					 Start the cooldown when we do.
				*/
				if (charController.isGrounded)
				{ 
					gustState = GustState.Wait;
					gustCooldownTimer.Start(gustCooldown);
				}
			} break;
			case GustState.Wait:
			{ 
				/*
					Waiting for the power to recharge, but unlike DashWait, 
					 this performs no move speed interpolation. GustState.Wait is 
					 set by the GustState.AirbornLockout case.
				*/
				float totalTime = 0f;
				if (gustCooldownTimer.UpdateAndCheck(ref totalTime))
				{ 
					gustState = GustState.Ready; 
					appliedMoveSpeed = baseMoveSpeed;
				}
			} break;
			default: {
				/* 
					If for some reason this manages to catch a state that it's not expecting, that's an 
					 error that a designer should be made aware of, because it should be impossible.
					Therefore, crash loudly.
				*/
				Debug.LogError("Switch statement on GustState has not handler for case: "+gustState.ToString()); 
				Debug.Break();
				Application.Quit();
			} break;
		}
	}
}
