﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using System.Reflection;

public static class Utilities
{
    /* It is frequently needed in various Start()s, to find and get a reference to other scripts.
        In many cases, those scripts MUST be found, or else the designer or programmer has made a
        mistake. In the case where the other component is not found, this script deliberately crashes
        to get the designer/programmer's attention. Otherwise, it fetches and sets the script via the 
        in/out ref variable component. */
    public static void RequireFindComponent<T>(ref T component, GameObject go) 
    where T : Component 
    {
        T t = go.GetComponent<T>();
        if (t == null)
        { 
            Debug.LogError("RequireFindComponent could not find "+typeof(T).ToString()+" on object "+go.name); 
            Debug.Break();
            Application.Quit();
        }
        component = t;
    }

    /* Just like AttachComponent, but calls GetComponentInChildren() instead of GetComponent(). */
    public static void RequireFindComponentInChildren<T>(ref T component, GameObject go) 
    where T : Component 
    {
        T t = go.GetComponentInChildren<T>();
        if (t == null)
        { 
            Debug.LogError("RequireFindComponentInChildren could not find "+typeof(T).ToString()+" on object "+go.name); 
            Debug.Break();
            Application.Quit();
        }
        component = t;
    }
}

public class Timer
{
    bool enabled = false;
    float startTime;
    float timerLength;
    public void Start(float time)
    {
        enabled = true;
        startTime = Time.time;
        timerLength = time;
    }

    public float Elapsed()
    { return Time.time - startTime; }

    public float Length()
    {return timerLength;}

    public void TurnOff()
    { enabled = false; }

    // returns true the first frame after the time on the timer has passed.
    //  otherwise, returns false.
    public bool UpdateAndCheck(ref float totalTimePassed)
    {
        if (enabled && Time.time - startTime > timerLength)
        {
            totalTimePassed = Time.time - startTime;
            enabled = false;
            return true;
        }
        else
        {
            return false;
        }
    }
    
    // returns true the first frame after the time on the timer has passed, also restarts timer.
    //  otherwise, returns false.
    public bool UpdateAndCheckLap(ref float totalTimePassed)
    {
        if (enabled && Time.time - startTime > timerLength)
        {
            totalTimePassed = Time.time - startTime;
            Start(timerLength);
            return true;
        }
        else
        {
            return false;
        }
    }
}